/*
 *  	int Forthish: "Everything's an int!"
 * 
 *		Virtual machine
 */

#include <stdio.h>
#include <allegro.h>
#include "vm.h"
#include "bytecode.h"


int stack[2048];
int pos = 0;

void push(int a) { stack[pos++] = a; }
int pop() { return stack[--pos]; }

/*--------------------------------------------------------------------*/

struct {
    char id[64];
    int value;	
} symbols[200];

int vm_symbol_value(char *id)
{
    int i = 0;
    
    while (symbols[i].id[0]) {
	if (strcmp(symbols[i].id, id) == 0)
	    return symbols[i].value;
	i++;
    }
    
    return -1;
}

/*
 * symbol must already exist in the table (i.e. be referenced)
 */
void vm_bind_symbol(char *id, int val)
{
    int i;
    
    for (i = 0; symbols[i].id[0]; i++) {
	if (strcmp(symbols[i].id, id) == 0) {
	    symbols[i].value = val;
	    break;
	}
    }
}

/*--------------------------------------------------------------------*/

void _put()
{
    int a, b;
    a = pop();
    b = pop();
    symbols[a].value = b;
}

void _ptrget() 
{
    int *p, off;
    off = pop();
    p = (int *) pop();
    push(*(p + off));
}

void _ptrput()
{
    int *p, off;
    off = pop();
    p = (int *) pop();
    *(p + off) = pop();
}

void _get() 
{
    push(symbols[pop()].value); 
}

void _dup()
{
    int a = pop();
    push(a); push(a);
}

void _drop()
{
    pop();
}

void _swap()
{
    int a, b;
    a = pop(); b = pop();
    push(a); push(b);
}

void _over()
{
    int x, y;
    y = pop(); x = pop();
    push(x); push(y); push(x);
}

void _plus()
{
    push(pop() + pop()); 
}

void _minus()
{
    int a, b;
    a = pop(); 
    b = pop();
    push(b - a);
}   

void _mult()
{
    push(pop() * pop()); 
}

void _div() 
{
    int a, b;
    a = pop();
    b = pop();
    push(b / a);
}

void _eq()
{
    push(pop() == pop()); 
}

void _lt() 
{
    int a, b;
    a = pop();
    b = pop();
    push(b < a);
}

void _lteq()
{
    int a, b;
    a = pop();
    b = pop();
    push(b <= a);
}

void _gt()
{
    int a, b;
    a = pop();
    b = pop();
    push(b > a);
}

void _gteq()
{
    int a, b;
    a = pop();
    b = pop();
    push(b >= a);
}

void _and()
{
    int a = pop(), b = pop();
    push(a && b); 
}

void _or()
{
    int a = pop(), b = pop();
    push(a || b); 
}

void _not()
{
    push(!pop()); 
}

void _jump(int *ip)
{
    *ip = pop() - 1;	/* ip is auto-incremented in vm_execute */
}

void _jumpf(int *ip) 
{
    int dest, test;
    dest = pop();
    test = pop();
    if (!test) 
	*ip = dest - 1;
}

void _jumpt(int *ip) 
{
    int dest, test;
    dest = pop();
    test = pop();
    if (test)
	*ip = dest - 1;
}

void _print()
{
    /* debug purposes only */
    int a = pop();
    printf("%d (%xh)\n", a, a);
}

void _call()
{
    int addr = pop();
    void (*p)(void);

    if (!addr) {
	printf("Call to unresolved function\n");
	return;
    }
    
    /* Sure, throw portability out the window... */
    if (addr & 0x80000000) {
	p = (void *)(addr & 0x7fffffff);
	p();
    }
    else
        vm_execute(addr);
}

/*--------------------------------------------------------------------*/

int vm_interrupt = 0;

int instr[4096];

void vm_execute(int pos)
{
    int ip, i;
    
    if (pos < 0)
	return;
    
    for (ip = pos; instr[ip] != RET; ip++) {
	i = instr[ip];
	
	if (i < 0) {
	    switch (i) {
		case PUT: _put(); break;
		case GET: _get();  break;
		case DUP: _dup(); break;
		case DROP: _drop(); break;
		case SWAP: _swap(); break;
		case OVER: _over(); break;
		
		case PTRPUT: _ptrput(); break;
		case PTRGET: _ptrget(); break;
		
		case PLUS: _plus(); break;
		case MINUS: _minus(); break;
		case MULT: _mult(); break;
		case DIV: _div(); break;
		
		case EQ: _eq(); break;
		case LT: _lt(); break;
		case LTEQ: _lteq(); break;
		case GT: _gt(); break;
		case GTEQ: _gteq(); break;

                case AND: _and(); break;
                case OR: _or(); break;
		case NOT: _not(); break;
		
		case JUMP: _jump(&ip); break;
		case JUMPF: _jumpf(&ip); break;
		case JUMPT: _jumpt(&ip); break;
		
		case CALL: _call(); break;
		
		case PRINT: _print(); break;

		case NEGNUM: push(-instr[++ip]); break;
		
		default:
		/* You are in heap big trouble, but I'll let you off 
		 * with a warning.  On second thought, screw it. */
	    }
	}
	else 
	    push(i);
	
	if (vm_interrupt)
	    return;
    }
}

/*--------------------------------------------------------------------*/

int vm_load_file(char *name)
{
    PACKFILE *f;
    int i = 0, c;
    
    f = pack_fopen(name, "rp");
    if (!f) return -1;
    
    /* instructions */
    while (1) {
	c = pack_igetw(f);
	if (c == EOF) goto verybad;
	
	c = (signed short) c;
	instr[i++] = c;
	if (c == NOMORE) 
	    break;
    }

    /* symbol table */
    for (i = 0; !pack_feof(f); i++) {
	pack_fgets(symbols[i].id, 64, f);
	symbols[i].value = (signed short) pack_igetw(f);
    }
    
    pack_fclose(f);
    return 0;
    
    verybad: {
	pack_fclose(f);
	return -1;
    }
}
